'''.
           @@@@@@@@@@
       @@@@..........@@@@
    @@@         .        @@@
  @@.           .         . @@
 @  .     _     .         .   @
@........| |...................@    *********************************************
@      . | |   _____  .        @
@      . | |  |  __ \ .        @    La Data Web
@      . | |__| |  | |.   ***  @
@........|____| |  | |...*   *.@    Copyright © 2022 Ignacio Barrau
@   .       . | |__| |. *     *@
@   .       . |_____/ . *     *@    *********************************************
@   .       .         . *     *@
@   .       .         . *******@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'''

import json
import requests
from simplepbi import utils
import pandas as pd

class Apps():
    """Simple library to use the Power BI api and obtain apps from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request app API
        Args:
            token: String
                Bearer Token to use the Power Bi Rest API
        ### Limitations
        ----
        Service principal authentication isn't supported.
        """
        self.token = token
        
    def get_app(self, app_id):
        """Returns the specified installed app.
        ### Parameters
        ----
        app_id: str uuid
            The Power Bi app id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing the specified installed app.
        ### Limitations
        ----
        Service principal authentication isn't supported.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/apps/{}".format(app_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_apps(self):
        """Returns a list of installed apps.
        ### Parameters
        ----
        None
        ### Returns
        ----
        Dict:
            A dictionary containing all the installed apps.
        ### Limitations
        ----
        Service principal authentication isn't supported.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/apps"
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
    
    def get_dashboard(self, app_id, dashboard_id):
        """Returns the specified dashboard from the specified app.
        ### Parameters
        ----
        app_id: str uuid
            The Power Bi app id. You can take it from PBI Service URL
        dashboard_id: str uuid
            The Power Bi dashboard id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing a dashboard in the specified app.
        ### Limitations
        ----
        Service principal authentication isn't supported.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/apps/{}/dashboards/{}".format(app_id, dashboard_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_dashboards(self, app_id):
        """Returns a list of dashboards from the specified app.
        ### Parameters
        ----
        app_id: str uuid
            The Power Bi app id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the dashboards in the specified app
        ### Limitations
        ----
        Service principal authentication isn't supported.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/apps/{}/dashboards".format(app_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_report(self, app_id, report_id):
        """Returns the specified report from the specified app.
        ### Parameters
        ----
        app_id: str uuid
            The Power Bi app id. You can take it from PBI Service URL
        report_id: str uuid
            The Power Bi report id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing a report in the specified app.
        ### Limitations
        ----
        Service principal authentication isn't supported.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/apps/{}/reports/{}".format(app_id, report_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_reports(self, app_id):
        """Returns a list of reports from the specified app.
        ### Parameters
        ----
        app_id: str uuid
            The Power Bi app id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the reports in the specified app
        ### Limitations
        ----
        Service principal authentication isn't supported.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/apps/{}/reports".format(app_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def get_tile(self, app_id, dashboard_id, tile_id):
        """Returns the specified tile within the specified dashboard from the specified app.
        Supported tiles include datasets and live tiles that contain an entire report page.
        ### Parameters
        ----
        app_id: str uuid
            The Power Bi app id. You can take it from PBI Service URL
        dashboard_id: str uuid
            The Power Bi dashboard id. You can take it from PBI Service URL
        tile_id: str uuid
            The tile id
        ### Returns
        ----
        Dict:
            A dictionary containing a tile in dashboard from the specified app.
        ### Limitations
        ----
        Service principal authentication isn't supported.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/apps/{appId}/dashboards/{}/tiles/{}".format(app_id, dashboard_id, tile_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
           
    def get_tiles(self, app_id, dashboard_id):
        """Returns a list of tiles within the specified dashboard from the specified app.
        ### Parameters
        ----
        app_id: str uuid
            The Power Bi app id. You can take it from PBI Service URL
        dashboard_id: str uuid
            The Power Bi dashboard id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the tiles in a dashboard from the specified app.
        ### Limitations
        ----
        Service principal authentication isn't supported.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/apps/{appId}/dashboards/{}/tiles".format(app_id, dashboard_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            



'''.
           @@@@@@@@@@
       @@@@..........@@@@
    @@@         .        @@@
  @@.           .         . @@
 @  .     _     .         .   @
@........| |...................@    *********************************************
@      . | |   _____  .        @
@      . | |  |  __ \ .        @    La Data Web
@      . | |__| |  | |.   ***  @
@........|____| |  | |...*   *.@    Copyright © 2022 Ignacio Barrau
@   .       . | |__| |. *     *@
@   .       . |_____/ . *     *@    *********************************************
@   .       .         . *     *@
@   .       .         . *******@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'''

import json
import requests
import io
import os
import base64
from simplepbi import utils
import pandas as pd
from requests_toolbelt.multipart.encoder import MultipartEncoder

class Imports():
    """Simple library to use the Power BI api and obtain imports from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request imports API
        Args:
            token: String
                Bearer Token to use the Power Bi Rest API
        """
        self.token = token
    
    def get_import(self, import_id):
        """Returns the specified import from My workspace.
        ### Parameters
        ----
        import_id: str uuid
            The Power Bi import id. 
        ### Returns
        ----
        Dict:
            A dictionary containing a import in My workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/imports/{}".format(import_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_import_in_group(self, workspace_id, import_id):
        """Returns the specified import from the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        import_id: str uuid
            The Power Bi import id.
        ### Returns
        ----
        Dict:
            A dictionary containing a import in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/imports/{}".format(workspace_id, import_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_imports(self):
        """Returns a list of imports from My workspace.
        ### Parameters
        ----
        None
        ### Returns
        ----
        Dict:
            A dictionary containing all the imports in My workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/imports"
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_imports_in_group(self, workspace_id):
        """Returns a list of imports from the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the imports in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/imports".format(workspace_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def simple_import_pbix(self, filePath, datasetDisplayName, nameConflict="CreateOrOverwrite", overrideModelLabel=None, overrideReportLabel=None):
        """Creates new content in My Workspace. Pbix with size lower than 1gb or with temporal blog storage url created
        Note: supported content for now Power BI .pbix files. Soon JSON files (.json), Excel files (.xlsx), SQL Server Report Definition Language files (.rdl)        
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        datasetDisplayName: str 
            The display name of the dataset should include file extension
        nameConflict: str
            Specifies what to do if a dataset with the same name already exists. The default value is Ignore. You can also use CreateOrOverwrite,GenerateUniqueName or Overwrite
        overrideModelLabel: str
            Determines whether to override the existing label on a model when republishing a Power BI .pbix file. The service default value is true.
        overrideReportLabel: str
            Whether to override the existing label on a report when republishing a Power BI .pbix file. The service default value is true.
        ### Request Body
        ----
        filePath: str
            Full local path like "C:/Users/SimplePBI/Documents/Filename.pbix"
        fileUrl
            SOON The shared access signature URL of the temporary blob storage used to import large Power BI .pbix files between 1 GB and 10 GB in size.
            
        ### Returns
        ----
        Dict:
            Response 202. A dict with a new report id.
        ### Limitations
        ----
        Importing a Power BI .pbix file from OneDrive isn't supported.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/imports?datasetDisplayName={}".format(datasetDisplayName)
            if nameConflict != None:
                url = url + "&nameConflict={}".format(str(nameConflict))
            if overrideModelLabel != None:
                url = url + "&overrideModelLabel={}".format(str(overrideModelLabel))
            if overrideReportLabel != None:
                url = url + "&overrideReportLabel={}".format(str(overrideReportLabel))                        
            # you need this dictionary to convert a binary file into form-data format
            # None here means we skip the filename and file content is important 
            files = {'value': (None, open(filePath, 'rb'), 'multipart/form-data')}
            # The MultipartEncoder is posted as data, don't use files=...!
            mp_encoder = MultipartEncoder(fields=files)
            # The MultipartEncoder provides the content-type header with the boundary:
            headers = {'Content-Type': 'multipart/form-data', "Authorization": "Bearer {}".format(self.token)}
            res = requests.post(url, data = mp_encoder, headers=headers)
            res.raise_for_status()
            
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def simple_import_pbix_in_group(self, workspace_id, filePath, datasetDisplayName, nameConflict="CreateOrOverwrite", overrideModelLabel=None, overrideReportLabel=None):
        """Creates new content in the specified workspace. Pbix with size lower than 1gb or with temporal blog storage url created
        Note: supported content for now Power BI .pbix files. Soon JSON files (.json), Excel files (.xlsx), SQL Server Report Definition Language files (.rdl)        
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        datasetDisplayName: str 
            The display name of the dataset should include file extension
        nameConflict: str
            Specifies what to do if a dataset with the same name already exists. The default value is Ignore. You can also use CreateOrOverwrite,GenerateUniqueName or Overwrite
        overrideModelLabel: str
            Determines whether to override the existing label on a model when republishing a Power BI .pbix file. The service default value is true.
        overrideReportLabel: str
            Whether to override the existing label on a report when republishing a Power BI .pbix file. The service default value is true.
        ### Request Body
        ----
        filePath: str
            Full local path like "C:/Users/SimplePBI/Documents/Filename.pbix"
        fileUrl
            SOON The shared access signature URL of the temporary blob storage used to import large Power BI .pbix files between 1 GB and 10 GB in size.
            
        ### Returns
        ----
        Dict:
            Response 202. A dict with a new report id.
        ### Limitations
        ----
        Importing a Power BI .pbix file from OneDrive isn't supported.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/imports?datasetDisplayName={}".format(workspace_id, datasetDisplayName)
            if nameConflict != None:
                url = url + "&nameConflict={}".format(str(nameConflict))
            if overrideModelLabel != None:
                url = url + "&overrideModelLabel={}".format(str(overrideModelLabel))
            if overrideReportLabel != None:
                url = url + "&overrideReportLabel={}".format(str(overrideReportLabel))                        
            # you need this dictionary to convert a binary file into form-data format
            # None here means we skip the filename and file content is important 
            files = {'value': (None, open(filePath, 'rb'), 'multipart/form-data')}
            # The MultipartEncoder is posted as data, don't use files=...!
            mp_encoder = MultipartEncoder(fields=files)
            # The MultipartEncoder provides the content-type header with the boundary:
            headers = {'Content-Type': 'multipart/form-data', "Authorization": "Bearer {}".format(self.token)}
            res = requests.post(url, data = mp_encoder, headers=headers)
            res.raise_for_status()
            
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def simple_import_pbix_as_parameter(self, workspace_id, fileContent, datasetDisplayName, nameConflict="CreateOrOverwrite", overrideModelLabel=None, overrideReportLabel=None):
        """Creates new content in the specified workspace from a Pbix added by parameter. Pbix with size lower than 1gb or with temporal blog storage url created
        Note: supported content for now Power BI .pbix files. Soon JSON files (.json), Excel files (.xlsx), SQL Server Report Definition Language files (.rdl)        
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        datasetDisplayName: str 
            The display name of the dataset should include file extension
        nameConflict: str
            Specifies what to do if a dataset with the same name already exists. The default value is Ignore. You can also use CreateOrOverwrite,GenerateUniqueName or Overwrite
        overrideModelLabel: str
            Determines whether to override the existing label on a model when republishing a Power BI .pbix file. The service default value is true.
        overrideReportLabel: str
            Whether to override the existing label on a report when republishing a Power BI .pbix file. The service default value is true.
        ### Request Body
        ----
        fileContent: str
            Pbix file as Response.Content from requests library. If you print the content it might end with something like this format: "x03\x00\x00I.\x10\x00\x00\x00"
        fileUrl
            SOON The shared access signature URL of the temporary blob storage used to import large Power BI .pbix files between 1 GB and 10 GB in size.
            
        ### Returns
        ----
        Dict:
            Response 202. A dict with a new report id.
        ### Limitations
        ----
        Importing a Power BI .pbix file from OneDrive isn't supported.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/imports?datasetDisplayName={}".format(workspace_id, datasetDisplayName)
            if nameConflict != None:
                url = url + "&nameConflict={}".format(str(nameConflict))
            if overrideModelLabel != None:
                url = url + "&overrideModelLabel={}".format(str(overrideModelLabel))
            if overrideReportLabel != None:
                url = url + "&overrideReportLabel={}".format(str(overrideReportLabel))                        
            # Convert Response.Content file to the correct format
            nube = io.BytesIO(fileContent)
            # you need this dictionary to convert a binary file into form-data format
            # None here means we skip the filename and file content is important 
            files = {'value': (None, nube, 'multipart/form-data')}
            # The MultipartEncoder is posted as data, don't use files=...!
            mp_encoder = MultipartEncoder(fields=files)
            # The MultipartEncoder provides the content-type header with the boundary:
            headers = {'Content-Type': 'multipart/form-data', "Authorization": "Bearer {}".format(self.token)}
            res = requests.post(url, data = mp_encoder, headers=headers)
            res.raise_for_status()            
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def simple_import_pbix_folder_preview(self, folderPath, nameConflict=None, overrideModelLabel=None, overrideReportLabel=None):
        """Creates new multiple contents in the specified workspace. Define a local folder with pbix files. Pbix with size lower than 1gb or with temporal blog storage url created
        Note: supported content for now Power BI .pbix files. Soon JSON files (.json), Excel files (.xlsx), SQL Server Report Definition Language files (.rdl)        
        ### Parameters
        ----
        datasetDisplayName: str 
            The display name of the dataset should include file extension            
        nameConflict: str
            Specifies what to do if a dataset with the same name already exists. The default value is Ignore. You can also use CreateOrOverwrite,GenerateUniqueName or Overwrite
        overrideModelLabel: str
            Determines whether to override the existing label on a model when republishing a Power BI .pbix file. The service default value is true.
        overrideReportLabel: str
            Whether to override the existing label on a report when republishing a Power BI .pbix file. The service default value is true.
        ### Request Body
        ----
        filePath: str
            Full local path of the files like "C:/Users/SimplePBI/Documents/"
        fileUrl: str
            SOON The shared access signature URL of the temporary blob storage used to import large Power BI .pbix files between 1 GB and 10 GB in size.
            
        ### Returns
        ----
        Dict:
            Response 202. A dict with a new report id.
        ### Limitations
        ----
        Importing a Power BI .pbix file from OneDrive isn't supported.
        """
        results = []
        try:
            allFiles= [files for root, dirs, files in os.walk(folderPath)][0]
            pbixFiles = [name for name in allFiles if name.endswith(("pbix"))]
            for datasetDisplayName in pbixFiles:
                url = "https://api.powerbi.com/v1.0/myorg/imports?datasetDisplayName={}".format(datasetDisplayName)
                if nameConflict != None:
                    url = url + "&nameConflict={}".format(str(nameConflict))
                if overrideModelLabel != None:
                    url = url + "&overrideModelLabel={}".format(str(overrideModelLabel))
                if overrideReportLabel != None:
                    url = url + "&overrideReportLabel={}".format(str(overrideReportLabel))                        
                # you need this dictionary to convert a binary file into form-data format
                # None here means we skip the filename and file content is important 
                files = {'value': (None, open(folderPath+datasetDisplayName, 'rb'), 'multipart/form-data')}
                # The MultipartEncoder is posted as data, don't use files=...!
                mp_encoder = MultipartEncoder(fields=files)
                # The MultipartEncoder provides the content-type header with the boundary:
                headers = {'Content-Type': 'multipart/form-data', "Authorization": "Bearer {}".format(self.token)}
                res = requests.post(url, data = mp_encoder, headers=headers)
                res.raise_for_status()
                results.append(res)
            return results
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def simple_import_pbix_folder_in_group_preview(self, workspace_id, folderPath, nameConflict=None, overrideModelLabel=None, overrideReportLabel=None):
        """Creates new multiple contents in the specified workspace. Define a local folder with pbix files. Pbix with size lower than 1gb or with temporal blog storage url created
        Note: supported content for now Power BI .pbix files. Soon JSON files (.json), Excel files (.xlsx), SQL Server Report Definition Language files (.rdl)        
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        folderPath: str 
            Set the file path from local file system like C:/Files/
        nameConflict: str
            Specifies what to do if a dataset with the same name already exists. The default value is Ignore. You can also use CreateOrOverwrite,GenerateUniqueName or Overwrite
        overrideModelLabel: str
            Determines whether to override the existing label on a model when republishing a Power BI .pbix file. The service default value is true.
        overrideReportLabel: str
            Whether to override the existing label on a report when republishing a Power BI .pbix file. The service default value is true.
        ### Request Body
        ----
        filePath: str
            Full local path of the files like "C:/Users/SimplePBI/Documents/"
        fileUrl: str
            SOON The shared access signature URL of the temporary blob storage used to import large Power BI .pbix files between 1 GB and 10 GB in size.
            
        ### Returns
        ----
        Dict:
            Response 202. A dict with a new report id.
        ### Limitations
        ----
        Importing a Power BI .pbix file from OneDrive isn't supported.
        """
        results = []
        try:
            allFiles= [files for root, dirs, files in os.walk(folderPath)][0]
            pbixFiles = [name for name in allFiles if name.endswith(("pbix"))]
            for datasetDisplayName in pbixFiles:
                url = "https://api.powerbi.com/v1.0/myorg/groups/{}/imports?datasetDisplayName={}".format(workspace_id, datasetDisplayName)
                if nameConflict != None:
                    url = url + "&nameConflict={}".format(str(nameConflict))
                if overrideModelLabel != None:
                    url = url + "&overrideModelLabel={}".format(str(overrideModelLabel))
                if overrideReportLabel != None:
                    url = url + "&overrideReportLabel={}".format(str(overrideReportLabel))                        
                # you need this dictionary to convert a binary file into form-data format
                # None here means we skip the filename and file content is important 
                files = {'value': (None, open(folderPath+datasetDisplayName, 'rb'), 'multipart/form-data')}
                # The MultipartEncoder is posted as data, don't use files=...!
                mp_encoder = MultipartEncoder(fields=files)
                # The MultipartEncoder provides the content-type header with the boundary:
                headers = {'Content-Type': 'multipart/form-data', "Authorization": "Bearer {}".format(self.token)}
                res = requests.post(url, data = mp_encoder, headers=headers)
                res.raise_for_status()
                results.append(res)
            return results
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def simple_import_from_devops(self, organization, project, repository_id, path, devopsKey, workspace_id):
        '''This function will import a PBIX file from DevOps to Power BI. Only available for Pbix with size lower than 1gb
         ### Parameters
        ----
        organization: str
            The organization name in DevOps. For example: https://dev.azure.com/ibarrau/ the organization name is ibarrau
        project: str
            The project name in DevOps. For example: https://dev.azure.com/ibarrau/ladataweb_proj/ the project name is ladataweb_proj
        repository_id: str
            The repository name in DevOps. Sometimes matches the name of the repo.
        path: str
            The path of the file in DevOps. For example: /Prod/PBITenantOverview.pbix
        devopsKey: str
            The DevOps Key. You can turn it on from DevOps > User Settings > Personal Access Tokens
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL    
        ### Returns
        ----
        Dict:
            Response 200 or 202 Ok
        '''
        file_name = path.split("/")[-1]
        try:
            pbix = requests.get(url="https://dev.azure.com/{}/{}/_apis/git/repositories/{}/items?path={}&download=true&api-version=6.0".format(organization, project, repository_id, path), auth=('user', devopsKey))
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
        try:
            res = self.simple_import_pbix_as_parameter(workspace_id, pbix.content, file_name)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def simple_import_from_github(self, owner, repo, path, github_pat, workspace_id):
        '''This function will import a PBIX file from GitHub to Power BI. The maximun size is 100Mb.
         ### Parameters
        ----
        owner: str
            The account owner of the repository. The name is not case sensitive. For example: https://dev.azure.com/ibarrau/ the organization name is ibarrau
        repo: str
            The name of the repository without the .git extension. The name is not case sensitive.
        path: str
            The path of the file in GitHub. For example: /Prod/PBITenantOverview.pbix
        github_pat: str
            The GitHub Personal Access Token. You can turn it on from Settings > Developer Settings > Personal Access Tokens
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL    
        ### Returns
        ----
        Dict:
            Response 202 Ok
        '''
        file_name = path.split("/")[-1]
        try:
            url = "https://api.github.com/repos/{}/{}/contents/{}".format(owner, repo, path)
            pbix_str = requests.get(url, headers={'Accept': 'application/vnd.github.raw+json', "Authorization": "Bearer {}".format(github_pat), 'X-GitHub-Api-Version': '2022-11-28' })
            #pbix_bytes = bytes(pbix_str.json()["content"], 'utf-8')
            #pbix = base64.decodebytes(pbix_bytes)
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
        try:
            res = self.simple_import_pbix_as_parameter(workspace_id, pbix_str.content, file_name)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)


'''.
           @@@@@@@@@@
       @@@@..........@@@@
    @@@         .        @@@
  @@.           .         . @@
 @  .     _     .         .   @
@........| |...................@    *********************************************
@      . | |   _____  .        @
@      . | |  |  __ \ .        @    La Data Web
@      . | |__| |  | |.   ***  @
@........|____| |  | |...*   *.@    Copyright © 2024 Ignacio Barrau
@   .       . | |__| |. *     *@
@   .       . |_____/ . *     *@    *********************************************
@   .       .         . *     *@
@   .       .         . *******@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'''

import json
import requests
from simplepbi import utils
import pandas as pd

class DataPipelines():
    """Simple library to use the  api and obtain data pipelines item from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request fabric data pipelines item API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token

    # Get Data Pipeline in Workspace
    def get_data_pipeline(self, workspace_id, data_pipeline_id):
        """Returns properties of the specified data pipeline.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        data_pipeline_id: str uuid
            The data pipeline id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing a data pipeline in the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/dataPipelines/{}".format(workspace_id, data_pipeline_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # List Data Pipelines in Workspace
    def list_data_pipelines(self, workspace_id):
        """Returns a list of data pipelines in the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the data pipelines in the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/dataPipelines".format(workspace_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Create Data Pipeline in Workspace
    def create_data_pipeline(self, workspace_id, displayName, description=None):
        """Creates a data pipeline in the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        displayName: str
            The data pipeline display name. The display name must follow naming rules according to item type.
        description: str
            The data pipeline description. Max length is 256 characters.
        ### Returns
        ----
        Response object from requests library. 201 or 202 OK
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/dataPipelines".format(workspace_id)
            body = {
                "displayName": displayName
            }
            if description != None:
                body["description"] = description
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.post(url, data=json.dumps(body), headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Delete Data Pipeline in Workspace
    def delete_data_pipeline(self, workspace_id, data_pipeline_id):
        """Deletes the specified data pipeline from the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        data_pipeline_id: str uuid
            The data pipeline id. You can take it from PBI Service URL
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/dataPipelines/{}".format(workspace_id, data_pipeline_id)
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.delete(url, headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
        
    # Update Data Pipeline in Workspace
    def update_data_pipeline(self, workspace_id, data_pipeline_id, displayName=None, description=None):
        """Updates the specified data pipeline in the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        data_pipeline_id: str uuid
            The data pipeline id. You can take it from PBI Service URL
        displayName: str
            The data pipeline display name. The display name must follow naming rules according to item type.
        description: str
            The data pipeline description. Max length is 256 characters.
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/dataPipelines/{}".format(workspace_id, data_pipeline_id)
            body = {}
            if displayName != None:
                body["displayName"] = displayName
            if description != None:
                body["description"] = description
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.patch(url, data=json.dumps(body), headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)


r'''.
           @@@@@@@@@@
       @@@@..........@@@@
    @@@         .        @@@
  @@.           .         . @@
 @  .     _     .         .   @
@........| |...................@    *********************************************
@      . | |   _____  .        @
@      . | |  |  __ \ .        @    La Data Web
@      . | |__| |  | |.   ***  @
@........|____| |  | |...*   *.@    Copyright © 2024 Ignacio Barrau
@   .       . | |__| |. *     *@
@   .       . |_____/ . *     *@    *********************************************
@   .       .         . *     *@
@   .       .         . *******@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'''

import json
import requests
from simplepbi import utils
import io
import pandas as pd

class Items():
    """Simple library to use the  api and obtain admin requests from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request fabric admin API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token
            
    def get_item(self, workspace_id, item_id, type):
        """Returns the specified item from the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        item_id: str uuid
            The item id. You can take it from PBI Service URL
        type: str 
            The type of the item. When querying for the following types, this parameter is required: { Report, Dashboard, SemanticModel, App, Dataflow }
        ### Returns
        ----
        Dict:
            A dictionary containing a item in the workspace.
        ### Limitations
        ----
        This API is supported for a number of item types, find the supported item types here.
        https://learn.microsoft.com/en-us/rest/api/fabric/articles/item-management/item-management-overview
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/workspaces/{}/items/{}?type={}".format(workspace_id, item_id, type)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
                      
    # List Items by workspace, capacity, type and status getting all paginated results from continuationToken in a single dictionary or pandas dataframe
    def list_items(self, workspace_id=None, capacity_id=None, type=None, status=None, return_pandas=False):
        """Returns a list of items for the specified workspace, capacity, type, and status.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        capacity_id: str uuid
            The capacity id.
        type: str
            The type of the item. When querying for the following types, this parameter is required: { Report, Dashboard, SemanticModel, App, Dataflow }
        status: str
            The status of the item. { "Active", "Deleted" }
        ### Returns
        ----
        Dict:
            A dictionary containing all the items.
        ### Limitations
        ----
        This API is supported for a number of item types, find the supported item types here.
        https://learn.microsoft.com/en-us/rest/api/fabric/articles/item-management/item-management-overview
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/items?"
            if workspace_id != None:
                url += "&workspaceId={}".format(workspace_id)
            if capacity_id != None:
                url += "&capacityId={}".format(capacity_id)
            if type != None:
                url += "&type={}".format(type)
            if status != None:
                url += "&status={}".format(status)
            url = url.replace("?&", "?")
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            data = res.json()
            while 'continuationToken' in data and data['continuationToken'] != None:
                url = "https://api.fabric.microsoft.com/v1/admin/items?continuationToken={}".format(data['continuationToken'])
                res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
                res.raise_for_status()
                data.update(res.json())
                data.pop('continuationToken')
            if return_pandas:
                js = json.dumps(data['value'])
                return pd.DataFrame(pd.read_json(io.StringIO(js)))
            else:
                return data
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Get List item Access Details by workspace and item of users
    def list_item_access_details(self, workspace_id, item_id, type=None):
        """Returns a list of users (including groups and service principals) and lists their workspace roles.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        item_id: str uuid
            The item id. You can take it from PBI Service URL
        type: str
            The type of the item. When querying for the following types, this parameter is required: { Report, Dashboard, SemanticModel, App, Dataflow }
        ### Returns
        ----
        Dict:
            A dictionary containing all the users who have access to the item.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/workspaces/{}/items/{}/users".format(workspace_id, item_id)
            if type != None:
                url += "?type={}".format(type)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)  
                
class Workspaces():
    """Simple library to use the api and obtain workspaces from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request fabric core workspaces API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token

    # Get Workspace
    def get_workspace(self, workspace_id):
        """Returns the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing a workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/workspaces/{}".format(workspace_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # List Workspaces by workspace role getting all paginated results from continuationToken in a single dictionary or pandas dataframe
    def list_workspaces(self, capacity_id=None, name=None, state=None, type=None, return_pandas=False):
        """Returns a list of workspaces
        ### Parameters
        ----
        capacity_id: str uuid
            The capacity id.
        name: str
            The name of the workspace.
        state: str
            The state of the workspace. { "Active", "Deleted" }
        type: str
            The type of the workspace. { "MyWorkspace", "WorkspaceV2" }
        ### Returns
        ----
        Dict:
            A dictionary containing all the workspaces.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/workspaces?"
            if capacity_id != None:
                url += "&capacityId={}".format(capacity_id)
            if name != None:
                url += "&name={}".format(name)
            if state != None:
                url += "&state={}".format(state)
            if type != None:
                url += "&type={}".format(type)
            url = url.replace("?&", "?")
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            data = res.json()
            while 'continuationToken' in data and data['continuationToken'] != None:
                url = "https://api.fabric.microsoft.com/v1/admin/workspaces?continuationToken={}".format(data['continuationToken'])
                res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
                res.raise_for_status()
                data.update(res.json())
                data.pop('continuationToken')
            if return_pandas:
                js = json.dumps(data['value'])
                return pd.DataFrame(pd.read_json(io.StringIO(js)))
            else:
                return data
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # List Workspace Access Details by workspace of users
    def list_workspace_access_details(self, workspace_id):
        """Returns a list of users (including groups and service principals) and lists their workspace roles.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the users who have access to the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/workspaces/{}/users".format(workspace_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

class Users():
    """Simple library to use the api and obtain users from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request fabric users API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token
        
    # List Access Entities by user and type getting all paginated results from continuationToken in a single dictionary or pandas dataframe
    def list_access_entities(self, user_id, type=None, return_pandas=False):
        """Returns a list of access entities for the specified user.
        ### Parameters
        ----
        user_id: str uuid
            The user id.
        type: str
            The type of the access entity. { "Report", "Datamart", "Lakehouse", "Dataflow", "Dashboard", "SemanticModel", "Notebook", etc }
        ### Returns
        ----
        Dict:
            A dictionary containing all the access entities.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/users/{}/access".format(user_id)
            if type != None:
                url += "?type={}".format(type)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            data = res.json()
            while 'continuationToken' in data and data['continuationToken'] != None:
                url = "https://api.fabric.microsoft.com/v1/admin/users/{}/access?continuationToken={}".format(user_id, data['continuationToken'])
                res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
                res.raise_for_status()
                data.update(res.json())
                data.pop('continuationToken')
            if return_pandas:
                js = json.dumps(data['value'])
                return pd.DataFrame(pd.read_json(io.StringIO(js)))
            else:
                return data
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

class Domains():
    """Simple library to use the api and obtain domains from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request domains API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token

    # Assign Domain's workspaces by capacity
    def assign_domains_workspaces(self, domain_id, capacities_id):
        """Assigns the specified domain to the specified capacity.
        ### Parameters
        ----
        domain_id: str uuid
            The domain id.
        capacities_id: str uuid[]
            List of capacities ids
        ### Returns
        ----
            A successful response.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains/{}/assignWorkspacesByCapacities".format(domain_id)
            body = {
                "capacitiesIds": capacities_id
            }
            res = requests.post(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Assign Domain Workspaces By list of Id
    def assign_domain_workspaces(self, domain_id, workspaces_id):
        """Assigns the specified domain to the specified workspaces.
        ### Parameters
        ----
        domain_id: str uuid
            The domain id.
        workspaces_id: str uuid[]
            List of workspaces id
        ### Returns
        ----
            A successful response.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains/{}/assignWorkspaces".format(domain_id)
            body = {
                "workspacesIds": workspaces_id
            }
            res = requests.post(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Assign Domain Workspaces By Principals
    def assign_domain_workspaces_principals(self, domain_id, principals):
        """Assigns the specified domain to the specified principals.
        ### Parameters
        ----
        domain_id: str uuid
            The domain id.
        principals: str[]
            The principals have two items: { "id": "string", "type": "string"}
        ### Returns
        ----
            A successful response.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains/{}/assignWorkspacesByPrincipals".format(domain_id)
            body = {
                "principals": principals
            }
            res = requests.post(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Create Domain
    def create_domain(self, display_name, description=None, parent_domain_id=None): 
        """Creates a domain.
        ### Parameters
        ----
        display_name: str
            The domain display name. The display name cannot contain more than 40 characters.
        description: str
            The domain description. The description cannot contain more than 256 characters.
        parent_domain_id: str uuid
            The domain parent object ID.
        ### Returns
        ----
            A successful response.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains"
            body = {
                "displayName": display_name
            }
            if description != None:
                body["description"] = description
            if parent_domain_id != None:
                body["parentDomainId"] = parent_domain_id
            res = requests.post(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}, data=json.dumps(body))
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Delete Domain
    def delete_domain(self, domain_id):
        """Deletes the specified domain.
        ### Parameters
        ----
        domain_id: str uuid
            The domain id.
        ### Returns
        ----
            A successful response.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains/{}".format(domain_id)
            res = requests.delete(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Get Domain
    def get_domain(self, domain_id):
        """Returns the specified domain.
        ### Parameters
        ----
        domain_id: str uuid
            The domain id.
        ### Returns
        ----
        Dict:
            A dictionary containing the domain.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains/{}".format(domain_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # List Domain Workspaces
    def list_domain_workspaces(self, domain_id):
        """Returns a list of workspaces for the specified domain.
        ### Parameters
        ----
        domain_id: str uuid
            The domain id.
        ### Returns
        ----
        Dict:
            A dictionary containing all the workspaces.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains/{}/workspaces".format(domain_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # List Domains 
    def list_domains(self, non_empty_only=None):
        """Returns a list of domains.
        ### Parameters
        ----
        non_empty_only: bool
            When true, only return domains that have at least one workspace that has an item in it. Default: false.
        ### Returns
        ----
        Dict:
            A dictionary containing all the domains.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains"
            if non_empty_only != None:
                url += "?nonEmptyOnly={}".format(non_empty_only)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Unassign Domain's workspaces by id
    def unassign_domain_workspaces(self, domain_id, workspaces_id):
        """Unassigns the specified domain from the specified workspaces.
        ### Parameters
        ----
        domain_id: str uuid
            The domain id.
        workspaces_id: str uuid[]
            List of workspaces id
        ### Returns
        ----
            A successful response.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains/{}/unassignWorkspaces".format(domain_id)
            body = {
                "workspacesIds": workspaces_id
            }
            res = requests.post(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Unassign All Domain workspaces
    def unassign_all_domain_workspaces(self, domain_id):
        """Unassigns the specified domain from all workspaces.
        ### Parameters
        ----
        domain_id: str uuid
            The domain id.
        ### Returns
        ----
            A successful response.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains/{}/unassignAllWorkspaces".format(domain_id)
            res = requests.post(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Update Domain
    def update_domain(self, domain_id, contributors_scope, display_name, description=None):
        """Updates the specified domain.
        ### Parameters
        ----
        domain_id: str uuid
            The domain id.
        contributors_scope: str
            The domain contributors scope. { "AdminsOnly", "AllTenant", "SpecificUsersAndGroups" }
        display_name: str
            The domain display name. The display name cannot contain more than 40 characters.
        description: str
            The domain description. The description cannot contain more than 256 characters.
        ### Returns
        ----
            A successful response.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains/{}".format(domain_id)
            body = {
                "contributorsScope": contributors_scope,
                "displayName": display_name
            }
            if description != None:
                body["description"] = description
            res = requests.patch(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}, data=json.dumps(body))
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Role Assignments bulk principals assign in domain
    def role_assignments_bulk_assign(self, domain_id, role, principals):
        """Assigns the specified role to the specified principals in the specified domain.
        ### Parameters
        ----
        domain_id: str uuid
            The domain id.
        role: str
            The role to assign to the principals.
        principals: str[]
            The principals have two items: [{"id": "string", "type": "string"}]. Type can be { "User", "Group", "ServicePrincipal" }
        ### Returns
        ----
            A successful response.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains/{}/roleAssignments/bulkAssign".format(domain_id)
            body = {
                "role": role,
                "principals": principals
            }
            res = requests.post(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}, data=json.dumps(body))
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Role Assignments bulk principals unassign in domain
    def role_assignments_bulk_unassign(self, domain_id, role, principals):
        """Unassigns the specified role from the specified principals in the specified domain.
        ### Parameters
        ----
        domain_id: str uuid
            The domain id.
        role: str
            The role to unassign from the principals.
        principals: str[]
            The principals have two items: [{"id": "string", "type": "string"}]. Type can be { "User", "Group", "ServicePrincipal" }
        ### Returns
        ----
            A successful response.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/admin/domains/{}/roleAssignments/bulkUnassign".format(domain_id)
            body = {
                "role": role,
                "principals": principals
            }
            res = requests.post(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}, data=json.dumps(body))
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)



'''.
           @@@@@@@@@@
       @@@@..........@@@@
    @@@         .        @@@
  @@.           .         . @@
 @  .     _     .         .   @
@........| |...................@    *********************************************
@      . | |   _____  .        @
@      . | |  |  __ \ .        @    La Data Web
@      . | |__| |  | |.   ***  @
@........|____| |  | |...*   *.@    Copyright © 2022 Ignacio Barrau
@   .       . | |__| |. *     *@
@   .       . |_____/ . *     *@    *********************************************
@   .       .         . *     *@
@   .       .         . *******@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'''

import json
import requests
from simplepbi import utils
import pandas as pd

class Capacities():
    """Simple library to use the Power BI api and obtain capacities from it. The user must have administrator rights or assign permissions on the capacity.
    """

    def __init__(self, token):
        """Create a simplePBI object to request capacities API. The user must have administrator rights or assign permissions on the capacity.
        Args:
            token: String
                Bearer Token to use the Power Bi Rest API
        """
        self.token = token
            
    def get_capacities(self):
        """Returns a list of capacities that the user has access to.
        ### Parameters
        ----
        None
        ### Returns
        ----
        Dict:
            A dictionary containing all the capacities in My workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/capacities"
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_refreshable_capacity(self, capacity_id, refreshable_id, expand=None):
        """Returns the specified refreshable for the specified capacity that the user has access to.
        Power BI retains a seven-day refresh history for each dataset, up to a maximum of sixty refreshes.
        ### Parameters
        ----
        capacity_id:
            The Power Bi capacity id. You can take it from PBI Service URL
        refreshable_id
            The Power Bi refreshable id in the capacity.
        expand: int
            Expands related entities inline, receives a comma-separated list of data types. Supported: capacities and groups
        ### Returns
        ----
        Dict:
            A dictionary containing all the refreshables in the capacity.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/capacities/{}/refreshables/{}?".format(capacity_id, refreshable_id)
            if expand != None:
                url = url + "$expand={}".format(expand)
            res = requests.get(url, headers={'Content-Type': 'capacitielication/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_refreshables_capacity(self, capacity_id, top, expand=None, filter=None, skip=None):
        """Returns a list of users that have access to the specified capacitie (Preview).
        ### Parameters
        ----
        capacity_id:
            The Power Bi capacitie id. You can take it from PBI Service URL
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        expand: string
            Expands related entities inline, receives a comma-separated list of data types. Supported: users, reports, dashboards, datasets, dataflows, workbooks
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the users in the capacity.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/capacities/{}/refreshables?$top={}".format(capacity_id, top) 
            if expand != None:
                url = url + "&$expand={}".format(expand)
            if filter != None:
                url = url + "&$filter={}".format(filter)
            if skip != None:
                url = url + "&$skip={}".format(skip) 
            res = requests.get(url, headers={'Content-Type': 'capacitielication/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_refreshables(self, top, expand=None, filter=None, skip=None):
        """Returns a list of refreshables for the organization.
        ### Parameters
        ----
        top: int
            Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
        expand: string
            Expands related entities inline, receives a comma-separated list of data types. Supported: users, reports, refreshables, datasets, dataflows, workbooks
        filter: string
            Filters the results based on a boolean condition
        skip: int 
            Skips the first n results. Use with top to fetch results beyond the first 5000.
        ### Returns
        ----
        Dict:
            A dictionary containing all the refreshables in the tenant.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/capacities/refreshables?$top={}".format(top)
            if expand != None:
                url = url + "&$expand={}".format(expand)
            if filter != None:
                url = url + "&$filter={}".format(filter)            
            if skip != None:
                url = url + "&$skip={}".format(skip)  
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_workload(self, capacity_id, workloadName):
        """Returns the current state of a workload. If the workload is enabled, the percentage of maximum memory that the workload can consume is also returned.
        Workload APIs aren't relevant for Embedded Gen2 capacities.
        ### Parameters
        ----
        capacity_id: str uuid
            The capacity id
        workloadName: str
            The name of the workload
        ### Returns
        ----
        Dict:
            A dictionary containing all the capacities in My workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/capacities/{}/Workloads/{}".format(capacity_id, workloadName)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_workloads(self, capacity_id):
        """Returns the current state of the specified capacity workloads. If a workload is enabled, the percentage of maximum memory that the workload can consume is also returned.
        ### Parameters
        ----
        capacity_id: str uuid
            The capacity id
        ### Returns
        ----
        Dict:
            A dictionary containing all the capacities in My workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/capacities/{}/Workloads".format(capacity_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def assign_myworkspaces_to_capacity(self, tagetCapacityObjectId):
        """Assigns My workspace to the specified capacity. To unassign My workspace from a capacity, provide an empty GUID (00000000-0000-0000-0000-000000000000) as the capacityId. This requires Admin rights or permissions on the capacity.
        ### Parameters
        ----
        ### Request Body
        ----
        targetCapacityObjectId: str uuid
            The premium capacity ID
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/AssignToCapacity"
            body ={
                "capacityId":tagetCapacityObjectId
            }
                
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def assign_to_capacity(self, tagetCapacityObjectId, workspacesToAssign):
        """Assigns the specified workspace to the specified capacity. To unassign My workspace from a capacity, provide an empty GUID (00000000-0000-0000-0000-000000000000) as the capacityId. This requires Admin rights or permissions on the capacity.
        ### Parameters
        ----
        ### Request Body
        ----
        targetCapacityObjectId: str uuid
            The premium capacity ID
        workspacesToAssign: str uuid
            List of the workspace IDs to be migrated to premium capacity
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/groups/{}/AssignToCapacity".format(workspacesToAssign)
            body ={
                "capacityId":tagetCapacityObjectId
            }
                
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_capacity_assignment_status(self):
        """Gets the status of the My workspace assignment-to-capacity operation.
        ### Parameters
        ----
        None
        ### Returns
        ----
        Dict:
            A dictionary containing all the CapacityAssignmentStatus in My workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/CapacityAssignmentStatus"
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_capacity_assignment_status_in_group(self, workspace_id):
        """Gets the status of the assignment-to-capacity operation for the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The Power Bi workspace id. You can take it from PBI Service URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the CapacityAssignmentStatus in the workspace.
        """
        try:
            url = "https://api.powerbi.com/v1.0/myorg/groups/{}/CapacityAssignmentStatus".format(workspace_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def patch_workload_preview(self, capacityId, workloadName, state, maxMemoryPercentageSetByUser=None ):    
        """Changes the state of a specific workload to Enabled or Disabled. When enabling a workload, specify the percentage of maximum memory that the workload can consume.
        *** THIS REQUEST IS IN PREVIEW IN SIMPLEPBI ***
        ### Parameters
        ----
        capacityId: str uud
            The capacity ID.
        workloadName: str
            The name of the workload
        ### Request Body
        ----
        state: str
            The capacity workload state. E.g { "Enabled, Disabled, Unsupported" }
        maxMemoryPercentageSetByUser: int
            The percentage of the maximum memory that a workload can consume (set by the user)
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.powerbi.com/v1.0/myorg/capacities/{}/Workloads/{}".format(capacityId, workloadName)
            body = {
                "state": state
            }            
            if maxMemoryPercentageSetByUser != None:
                body["maxMemoryPercentageSetByUser"]=maxMemoryPercentageSetByUser
                
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            
            res = requests.patch(url, json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)


'''.
           @@@@@@@@@@
       @@@@..........@@@@
    @@@         .        @@@
  @@.           .         . @@
 @  .     _     .         .   @
@........| |...................@    *********************************************
@      . | |   _____  .        @
@      . | |  |  __ \ .        @    La Data Web
@      . | |__| |  | |.   ***  @
@........|____| |  | |...*   *.@    Copyright © 2022 Ignacio Barrau
@   .       . | |__| |. *     *@
@   .       . |_____/ . *     *@    *********************************************
@   .       .         . *     *@
@   .       .         . *******@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'''

import json
import requests

class Token():
    """Simple library to use the Power BI api and obtain datasets from it.
    """

    def __init__(self, tenant_id=None, power_bi_client_id=None, power_bi_username=None, power_bi_password=None, power_bi_secret=None, use_service_principal=False):
        """Create a simplePBI object to get authentication token.
        
        You can use:
            1) User and pass authentication
                Provide:    tenant_id
                            power_bi_client_id
                            power_bi_username
                            power_bi_password
            2) Service principal authentication (set use_service_principal to True)
                Provide:    tenant_id
                            power_bi_client_id
                            power_bi_secret
                            use_service_principal=True
        Args:
            tenant_id : String
                Tenant ID to connect to.
            power_bi_client_id : String
                Client ID (also known as App ID)
            power_bi_username : String
                Username to use when authenticating.
            power_bi_password : String
                Password to use when authenticating.
            power_bi_secret : String
                The secret to authenticate with the Client ID.
            use_service_principal : Bool
                False by default, if set to true it will authenticate using power_bi_client_id and power_bi_secret.
        """

        POWER_BI_RESOURCE_ENDPOINT = "https://analysis.windows.net/powerbi/api"
        MICROSOFT_OAUTH2_API_ENDPOINT = "https://login.microsoftonline.com/" + tenant_id + "/oauth2/token/"
        if not use_service_principal:
            try:
                url = MICROSOFT_OAUTH2_API_ENDPOINT
                body = {
                    "resource":POWER_BI_RESOURCE_ENDPOINT, 
                    "client_id":power_bi_client_id,
                    "grant_type":"password",
                    "username":power_bi_username,
                    "password":power_bi_password,
                    "scope":"openid"                    
                }
                headers = {
                    "Content-Type": "application/x-www-form-urlencoded"
                    }
                r = requests.post(url = url, data = body, headers = headers)
                access_token = r.json().get('access_token')
                self.token = access_token
            except requests.exceptions.HTTPError as ex:
                print(ex)
            except Exception as e:
                print(e)
        else:
            try:
                url = MICROSOFT_OAUTH2_API_ENDPOINT #is this different?
                body = {
                    "grant_type":"client_credentials",
                    "client_id":power_bi_client_id,
                    "client_secret":power_bi_secret,
                    "resource":POWER_BI_RESOURCE_ENDPOINT            
                }
                headers = {
                    "Content-Type": "application/x-www-form-urlencoded"
                    }
                r = requests.post(url = url, data = body, headers = headers)
                access_token = r.json().get('access_token')
                self.token = access_token
            except requests.exceptions.HTTPError as ex:
                print(ex)
    
    


r'''.
           @@@@@@@@@@
       @@@@..........@@@@
    @@@         .        @@@
  @@.           .         . @@
 @  .     _     .         .   @
@........| |...................@    *********************************************
@      . | |   _____  .        @
@      . | |  |  __ \ .        @    La Data Web
@      . | |__| |  | |.   ***  @
@........|____| |  | |...*   *.@    Copyright © 2024 Ignacio Barrau
@   .       . | |__| |. *     *@
@   .       . |_____/ . *     *@    *********************************************
@   .       .         . *     *@
@   .       .         . *******@
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'''
import json
import requests
from simplepbi import utils
import pandas as pd
import os
import io
import base64

class Items():
    """Simple library to use the  api and obtain items from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request fabric item API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token
            
    def get_item(self, workspace_id, item_id):
        """Returns the specified item from the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL
        ### Returns
        ----
        Dict:
            A dictionary containing a item in the workspace.
        ### Limitations
        ----
        This API is supported for a number of item types, find the supported item types here.
        https://learn.microsoft.com/en-us/rest/api/fabric/articles/item-management/item-management-overview
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}".format(workspace_id, item_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
                      
    def list_items(self, workspace_id, return_pandas=False, type=None):
        """Returns a list of items from the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        return_pandas: bool
            If True, returns a pandas DataFrame with the items. Default is False, returns a dictionary. 
        type: str
            The type of items to return. If None, returns all item types.
        ### Returns
        ----
        Dict:
            A dictionary containing all the items in the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/items".format(workspace_id)
            if type != None:
                url += "?type={}".format(type)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            data = res.json()
            while 'continuationToken' in data and data['continuationToken'] != None:
                url = "https://api.fabric.microsoft.com/v1/workspaces/{}/items?continuationToken={}".format(workspace_id, data['continuationToken'])
                if type != None:
                    url += "&type={}".format(type)
                res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
                res.raise_for_status()
                data.update(res.json())
                data.pop('continuationToken')
            if return_pandas:
                js = json.dumps(data['value'])
                return pd.DataFrame(pd.read_json(io.StringIO(js)))
            else:
                return data            
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
                       
    def delete_item(self, workspace_id, item_id):
        """Deletes the specified item from the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL
        ### Returns
        ----
        Response object from requests library. 200 OK
        
        """
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}".format(workspace_id, item_id)   
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.delete(url, headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def create_item(self, workspace_id, displayName, itemType, description=None, parts=None):
        """Creates an item in the specified workspace. Preview request, soon we'll add 'definition' parameter
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        displayName: str
            The item display name. The display name must follow naming rules according to item type.
        itemType: str
            The item type. Example: Dashboard, DataPipeline, Datamart, Eventstream, KQLDataConnection, KQLDatabase, KQLQueryset, Lakehouse, MLExperiment, MLModel, MountedWarehouse, Notebook, PaginatedReport, Report, SQLEndpoint, SemanticModel, SparkJobDefinition
        description: str
            The item description. Max length is 256 characters.
        ### Returns
        ----
        Response object from requests library. 201 or 202 OK
        ### Limitations
        ----
        To create a non-PowerBI Fabric item the workspace must be on a supported Fabric capacity type. For more information see Microsoft Fabric license types.
To create a PowerBI item, the user must have the appropritate license. For more information see Microsoft Fabric license types.
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/items".format(workspace_id)
            body = {
                "displayName": displayName,
                "type": itemType
            }
            if description != None:
                body["description"]=description
            if parts != None:
                body["definition"]={ "Parts": parts }
            headers={'Content-Type': 'application/json; charset=utf-8', "Authorization": "Bearer {}".format(self.token)}
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            if res.status_code==202:
                print("Request accepted, item provisioning in progress. Please wait. Operation id: ", res.headers['x-ms-operation-id'])
                # Get operation state
                LongRunningOperations(self.token).get_operation_state(res.headers['x-ms-operation-id'])                
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_item_definition(self, workspace_id, item_id):
        """Returns the specified item definition.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL        
        ### Returns
        ----
        Response object from requests library. 200 OK        
        ### Limitations
        ----
        This API is blocked for an item with an encrypted sensitivity label.
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}/getDefinition".format(workspace_id, item_id)
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def update_item(self, workspace_id, item_id, displayName=None, description=None):
        """Updates the properties of the specified item.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL
        displayName: str
            The item display name. The display name must follow naming rules according to item type.
        description: str
            The item description. Max length is 256 characters.
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}".format(workspace_id, item_id)
            body = {}
            if displayName != None:
                body["displayName"]=displayName
            if description != None:
                body["description"]=description
            if body == {}:
                raise Exception("Please specify a display name or description to update item.")
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.patch(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def update_item_definition(self, workspace_id, item_id, parts, format=None):
        """Overrides the definition for the specified item.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL
        parts: ItemDefinitionPart[]
            A list of definition parts. part, payload and payloadtype description. Read API Docs for more details.
        format: str 
            	The format of the item definition.
        
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}/updateDefinition".format(workspace_id, item_id)
            body = {
                "definition": {
                    "Parts": parts
                }
            }
            if format != None:
                body["format"]=format
            headers={'Content-Type': 'application/json; charset=utf-8', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()            
            if res.status_code==202:
                print("Request accepted, item provisioning in progress. Please wait. Operation id: ", res.headers['x-ms-operation-id'])
                # Get operation state
                LongRunningOperations(self.token).get_operation_state(res.headers['x-ms-operation-id'])                
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def build_semantic_model_parts(self, item_path):
        """Build the parts of semantic model for the specified item.
        #### Parameters
        ----
        item_path: str 
            The semantic model path until [name].Report folder like C:/Users/user/Desktop/[name].SemanticModel or .Dataset
        ### Returns
        ----
        Dict with parts of the semantic model
        """ 
        parts = []        

        # Iterate through the files in the item_path
        for root, dirs, files in os.walk(item_path):
            if os.path.basename(root) == ".pbi":
                continue
            for file in files:
                # Skip files with the name "item.*.json"
                if file.startswith("item.") and file.endswith(".json"):
                    continue
                if file == "cache.abf":
                    continue

                # Get the file path relative to the project folder
                file_path = os.path.relpath(os.path.join(root, file), item_path).replace("\\","/")

                # Read the file contents
                with open(os.path.join(root, file), "rb") as f:
                    file_contents = f.read()

                # Base64-encode the file contents
                encoded_contents = base64.b64encode(file_contents).decode("utf-8")

                # Add the file to the Parts array
                parts.append({
                    "Path": file_path,
                    "Payload": encoded_contents,
                    "PayloadType": "InlineBase64"
                })
        return parts

    def simple_deploy_semantic_model(self, workspace_id, item_path):
        """Deploys the semantic model for the specified item.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_path: str 
            The semantic model path until [name].SemanticModel folder like C:/Users/user/Desktop/[name].SemanticModel
        ### Returns
        ----
        Response object from requests library. 202 OK
        """
        
        item_name = item_path.split("/")[-1].split(".")[0]
        if item_name == "":
            raise Exception("Make sure the path doesn't en in / or \\ at the end.")
        else:
            print("Item name: ", item_name)
        parts = self.build_semantic_model_parts(item_path)

        try:
            # Get list of items
            it = self.list_items(workspace_id)
            id = [i['id'] for i in it['value'] if i['displayName']==item_name and i['type']=="SemanticModel" ]
            if id == []:
                # Create item
                res = self.create_item(workspace_id, item_name, "SemanticModel", None, parts)
            else:
                # Update item definition
                print("Updating semantic model id {} in workspace id {}".format(id[0], workspace_id))
                res = self.update_item_definition(workspace_id, id[0], parts)            
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def build_report_parts(self, semantic_model_workspace_id, item_path):
        """Build the parts of report for the specified item.
        #### Parameters
        ----
        semantic_model_workspace_id: str uuid
            The workspace id of the semantic model of a report. You can take it from Fabric URL
        item_path: str 
            The semantic model path until [name].Report folder like C:/Users/user/Desktop/[name].Report
        ### Returns
        ----
        Dict with parts of the report
        """     
        
        parts = []
        
        # Iterate through the files in the item_path
        for root, dirs, files in os.walk(item_path):
            # Skip folders with the name ".pbi"
            if os.path.basename(root) == ".pbi":
                continue

            for file in files:
                # Skip files with the name "item.*.json"
                if file.startswith("item.") and file.endswith(".json"):
                    continue
                if file == "cache.abf":
                    continue
                if file.endswith(".pbir"):
                    # Load the JSON file
                    with open(item_path +'/definition.pbir', 'r') as f:
                        pbir_json = json.load(f)
                        
                    # Remove the "byPath" item
                    semantic_model_name = pbir_json['datasetReference']['byPath']['path'].split("/")[-1].split(".")[0]                    
                    print("Looking for id of semantic model {} in workspace id {} related to the report".format(semantic_model_name, semantic_model_workspace_id))
                    try:
                        it = self.list_items(semantic_model_workspace_id)
                        semantic_model_id = [i['id'] for i in it['value'] if i['displayName']==semantic_model_name and i['type']=="SemanticModel" ]
                        if semantic_model_id == []:
                            raise Exception("Semantic Model {} does not exist in the specified workspace.".format(semantic_model_name))
                        print("Semantic model id found: {}".format(semantic_model_id))
                    except Exception as e:
                        print("Error: ", e)
                        
                    del pbir_json['datasetReference']['byPath']
                    del pbir_json["$schema"]
                    
                    # Add a new JSON object to the "byConnection" property
                    pbir_json['datasetReference']['byConnection'] = {
                        "connectionString": None,
                        "pbiServiceModelId": None,
                        "pbiModelVirtualServerName": "sobe_wowvirtualserver",
                        "pbiModelDatabaseName": semantic_model_id[0],
                        "name": "EntityDataSource",
                        "connectionType": "pbiServiceXmlaStyleLive"
                    }
                    # Convert the PBIR JSON object to a string
                    pbir_json_str = json.dumps(pbir_json)
                    
                    # Convert the string to UTF-8 bytes
                    file_contents = pbir_json_str.encode('utf-8')
                else:            
                    # Read the file contents
                    with open(os.path.join(root, file), "rb") as f:
                        file_contents = f.read()
                
                # Get the file path relative to the project folder
                file_path = os.path.relpath(os.path.join(root, file), item_path).replace("\\","/")
            
                # Base64-encode the file contents
                encoded_contents = base64.b64encode(file_contents).decode("utf-8")

                # Add the file to the Parts array
                parts.append({
                    "Path": file_path,
                    "Payload": encoded_contents,
                    "PayloadType": "InlineBase64"
                })
        return parts

    def simple_deploy_report(self, report_workspace_id, semantic_model_workspace_id, item_path):
        """Deploys the semantic model for the specified item.
        #### Parameters
        ----
        report_workspace_id: str uuid
            The workspace id of the destination of the report deployment. You can take it from Fabric URL
        item_path: str 
            The semantic model path until [name].Report folder like C:/Users/user/Desktop/[name].Report
        ### Returns
        ----
        Response object from requests library. 202 OK
        """       

        item_name = item_path.split("/")[-1].split(".")[0]
        if item_name == "":
            raise Exception("Make sure the path doesn't en in / or \\ at the end.")
        else:
            print("Item name: ", item_name)

        parts = self.build_report_parts(semantic_model_workspace_id, item_path)

        try:
            # Get list of items
            it = self.list_items(report_workspace_id)
            id = [i['id'] for i in it['value'] if i['displayName']==item_name and i['type']=="Report" ]
            if id == []:
                # Create item
                res = self.create_item(report_workspace_id, item_name, "Report", None, parts)
            else:
                # Update item definition
                print("Updating report id {} in workspace id {}".format(id[0], report_workspace_id))
                res = self.update_item_definition(report_workspace_id, id[0], parts)            
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
                
       
class Git():
    """Simple library to use the  api and obtain items from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request fabric item API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token
            
    def get_git_connection(self, workspace_id):
        """Returns git connection details for the specified workspace.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        ### Returns
        ----
        Dict:
            A dictionary containing a item in the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/git/connection".format(workspace_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def get_git_status(self, workspace_id):
        """Returns the Git status of items in the workspace, that can be committed to Git.
        ### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        ### Returns
        ----
        Dict:
            A dictionary containing a item in the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/git/status".format(workspace_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def commit_to_git(self, workspace_id, mode, comment, workspaceHead, items=None):
        """Commits the changes made in the workspace to the connected remote branch.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        mode: string
            The mode for the commit operation.
        comment: string
            	Caller-free comment for this commit. Maximum length is 300 characters. If no comment is provided by the caller, use the default Git provider comment.
        workspaceHead: str 
            	Full SHA hash that the workspace is synced to. The hash can be retrieved from the Git Status API.
        items: itemsIdentifier[]
            Specific items to commit. This is relevant only for Selective commit mode. The items can be retrieved from the Git Status API.
            Each item it's { "logicalId": str uuid, "objectId": str uud }. You can use one of them or both if you have them.
        ### Returns
        ----
        Response object from requests library. 202 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/git/commitToGit".format(workspace_id)
            body = {
                "mode": mode,
                "workspaceHead": workspaceHead,
                "comment": comment
            }
            if items != None:
                body["items"]=items
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
    def update_from_git(self, workspace_id, item_id, remoteCommitHash, conflictResolution, allowOverrideItems, workspaceHead):
        """Updates the workspace with commits pushed to the connected branch.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL
        remoteCommitHash: string
            	Remote full SHA commit hash.
        conflictResolution: string
            	Conflict resolution to be used in the update from Git operation. If items are in conflict and a conflict resolution is not specified, the update operation will not start. Example:
            {"ConflictResolutionPolicy": PreferRemote or PreferWorkspace, "ConflictResolutionType": "Workspace"}
        options: options
            Automatically written by simplepbi. Options to be used in the update from Git operation. For now just overrideitems
        allowOverrideItems: bool
            User consent to override incoming items during the update from Git process. When incoming items are present and the allow override items is not specified or is provided as false, the update operation will not start. Default value is false.
        workspace_head: str 
            	Full SHA hash that the workspace is synced to. The hash can be retrieved from the Git Status API.
        ### Returns
        ----
        Response object from requests library. 202 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}/git/updateFromGit".format(workspace_id, item_id)
            body = {
                "workspaceHead": workspaceHead,
                "remoteCommitHash": remoteCommitHash,
                "conflictResolution": conflictResolution,
                "options":{
                    "allowOverrideItems": allowOverrideItems
                }
            }
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Connect to git repo
    def connect_git_repo(self, workspace_id, organization, project, repository, branch, directory_name, git_provider_type):
        """Connects the workspace to an Azure DevOps repository.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        organization: string
            The Azure DevOps organization name.
        project: string
            The Azure DevOps project name.
        repository: string
            The Azure DevOps repository name.
        branch: string
            The Azure DevOps branch name.
        directory_name: string
            The Azure DevOps directory name.
        git_provider_type: string
            The Git provider type. Example: AzureDevOps
        ### Returns
        ----
        Response object from requests library. 202 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/git/connect".format(workspace_id)
            body = {
                "gitProviderDetails": {
                    "organization": organization,
                    "project": project,
                    "repository": repository,
                    "branch": branch,
                    "directoryName": directory_name,
                    "gitProviderType": git_provider_type
                    }
            }
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Disconnect from git repo
    def disconnect_git_repo(self, workspace_id):
        """Disconnects the workspace from an Azure DevOps repository.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        ### Returns
        ----
        Response object from requests library. 202 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/git/disconnect".format(workspace_id)
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
                    
            
class Scheduler():

    """Simple library to use the api and obtain job scheduler from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request fabric job scheduler API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token
        
    def run_on_demand_item_job(self, workspace_id, item_id, jobType):
        """Run on-demand item job instance.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL
        jobType: string
            The job type, for now just "DefaultJob"
        executionData: str 
            	Execution data for the job. Preview parameter, a working progress from the API      
        ### Returns
        ----
        Response object from requests library. 202 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}/jobs/instances?jobType={}".format(workspace_id, item_id, jobType)
            body = {
                "executionData": {}
            }
            if format != None:
                body["format"]=format
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Get Item Job Instance
    def get_item_job_instance(self, workspace_id, item_id, job_instance_id):
        """Returns the specified job instance for the specified item.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL
        job_instance_id: str uuid
            The job instance id.
        ### Returns
        ----
        Dict:
            A dictionary containing a job instance in the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}/jobs/instances/{}".format(workspace_id, item_id, job_instance_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Cancel Item Job Instance
    def cancel_item_job_instance(self, workspace_id, item_id, job_instance_id):
        """Cancels the specified job instance for the specified item.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL
        job_instance_id: str uuid
            The job instance id.
        ### Returns
        ----
        Response object from requests library. 202 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}/jobs/instances/{}".format(workspace_id, item_id, job_instance_id)
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
            
class Workspaces():
    """Simple library to use the api and obtain workspaces from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request fabric core workspaces API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token

    # Add Workspace Role Assignment
    def add_workspace_role_assignment(self, workspace_id, principal_id, principal_type, role):
        """Adds a role assignment to the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        principal_id: str uuid
            The principal id.
        principal_type: string
            The principal type. { "User", "Group", "ServicePrincipal", "ServicePrincipalProfile" }
        role: string
            The role. Workspace role, like { "Member", "Admin", "Contributor", "Viewer"}
        ### Returns
        ----
        Response object from requests library. 202 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/roleAssignments".format(workspace_id)
            body = {
                "principal": {
                    "id": principal_id,
                    "type": principal_type,
                },
                "role": role
            }
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Assign Workspace to Capacity
    def assign_workspace_to_capacity(self, workspace_id, capacity_id):
        """Assigns the specified workspace to the specified capacity.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        capacity_id: str uuid
            The capacity id.
        ### Returns
        ----
        Response object from requests library. 202 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/assignToCapacity".format(workspace_id)
            body = {
                "capacityId": capacity_id
            }
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Create Workspace
    def create_workspace(self, displayName, capacity_id=None, description=None):
        """Creates a workspace.
        #### Parameters
        ----
        displayName: string
            The workspace display name. The display name must follow naming rules according to item type.
        capacity_id: str uuid
            The capacity id.
        description: string
            The workspace description. Max length is 256 characters.
        ### Returns
        ----
        Response object from requests library. 201 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces"
            body = {
                "displayName": displayName
            }
            if capacity_id != None:
                body["capacityId"]=capacity_id
            if description != None:
                body["description"]=description
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Delete Workspace
    def delete_workspace(self, workspace_id):
        """Deletes the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}".format(workspace_id)
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.delete(url, headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Delete Workspace Role Assignment by principal id
    def delete_workspace_role_assignment_by_principal_id(self, workspace_id, principal_id):
        """Deletes the specified role assignment for the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        principal_id: str uuid
            The principal id.
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/roleAssignments/{}".format(workspace_id, principal_id)
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.delete(url, headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Get Workspace
    def get_workspace(self, workspace_id):
        """Returns the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        ### Returns
        ----
        Dict:
            A dictionary containing a workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}".format(workspace_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # List Workspaces by workspace role getting all paginated results from continuationToken in a single dictionary or pandas dataframe
    def list_workspaces(self, return_pandas=False, roles=None):
        """Returns a list of workspaces for the specified role.
        #### Parameters
        ----
        roles: string
            The role. Workspace role, like { "Member", "Admin", "Contributor", "Viewer"}
        ### Returns
        ----
        Dict:
            A dictionary containing all the workspaces.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces"
            if roles != None:
                url = "https://api.fabric.microsoft.com/v1/workspaces?role={}".format(roles)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            data = res.json()
            while 'continuationToken' in data and data['continuationToken'] != None:
                url = "https://api.fabric.microsoft.com/v1/workspaces?role={}&continuationToken={}".format(roles, data['continuationToken'])
                res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
                res.raise_for_status()
                data.update(res.json())
                data.pop('continuationToken')
            if return_pandas:
                js = json.dumps(data['value'])
                return pd.DataFrame(pd.read_json(io.StringIO(js)))
            else:
                return data
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)
        
    # Unassign Workspace from Capacity
    def unassign_workspace_from_capacity(self, workspace_id):
        """Unassigns the specified workspace from the capacity.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/unassignFromCapacity".format(workspace_id)
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Update Workspace
    def update_workspace(self, workspace_id, displayName=None, description=None):
        """Updates the properties of the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        displayName: string
            The workspace display name. The display name must follow naming rules according to item type.
        description: string
            The workspace description. Max length is 256 characters.
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}".format(workspace_id)
            body = {}
            if displayName != None:
                body["displayName"]=displayName
            if description != None:
                body["description"]=description
            if body == {}:
                raise Exception("Please specify a display name or description to update workspace.")
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.patch(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Update Workspace Role Assignment
    def update_workspace_role_assignment(self, workspace_id, principal_id, role):
        """Updates the specified role assignment for the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        principal_id: str uuid
            The principal id.
        role: string
            The role. Workspace role, like { "Member", "Admin", "Contributor", "Viewer"}
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/roleAssignments/{}".format(workspace_id, principal_id)
            body = {
                "role": role
            }
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.patch(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

class Onelake():
    """Simple library to use the onelake api and obtain items from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request onelake API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token
    
    # Create shortcut
    def create_shortcut(self, workspace_id, item_id, name, path, target, shortcut_conflict_policy=None):
        """Creates a shortcut in the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL
        name: string
            The shortcut name.
        path: string
            The shortcut path.
        target: string
            The shortcut target.
        shortcut_conflict_policy: string
            The shortcut conflict policy. { "Abort", "GenerateUniqueName" }
        ### Returns
        ----
        Response object from requests library. 201 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}/shortcuts".format(workspace_id, item_id)
            if shortcut_conflict_policy != None:                
                url = url + "?shortcutConflictPolicy={}".format(shortcut_conflict_policy)
            body = {
                "name": name,
                "path": path,
                "target": target
            }            
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.post(url, data = json.dumps(body), headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Delete Shortcut
    def delete_shortcut(self, workspace_id, item_id, shortcut_path, shortcut_name):
        """Deletes the specified shortcut from the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL
        shortcut_path: string
            The shortcut path.
        shortcut_name: string
            The shortcut name.
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        
        try: 
            url= "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}/shortcuts/{}/{}".format(workspace_id, item_id, shortcut_path, shortcut_name)
            headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}            
            res = requests.delete(url, headers = headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # Get Shortcut
    def get_shortcut(self, workspace_id, item_id, shortcut_path, shortcut_name):
        """Returns the specified shortcut from the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL
        shortcut_path: string
            The shortcut path.
        shortcut_name: string
            The shortcut name.
        ### Returns
        ----
        Dict:
            A dictionary containing a shortcut in the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}/shortcuts/{}/{}".format(workspace_id, item_id, shortcut_path, shortcut_name)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    # List Data Access Roles in workspace in item getting all paginated results from continuationToken in a single dictionary or pandas dataframe
    def list_data_access_roles(self, workspace_id, item_id, return_pandas=False):
        """Returns a list of data access roles for the specified item in the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        item_id: str uuid
            The item id. You can take it from Fabric URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the data access roles in the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}/dataAccessRoles".format(workspace_id, item_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            data = res.json()
            while 'continuationToken' in data and data['continuationToken'] != None:
                url = "https://api.fabric.microsoft.com/v1/workspaces/{}/items/{}/dataAccessRoles?continuationToken={}".format(workspace_id, item_id, data['continuationToken'])
                res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
                res.raise_for_status()
                data.update(res.json())
                data.pop('continuationToken')
            if return_pandas:
                js = json.dumps(data['value'])
                return pd.DataFrame(pd.read_json(io.StringIO(js)))
            else:
                return data
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

class LongRunningOperations():
    """Simple library to use the Long Running Operations api and obtain operation status from it.
    """

    def __init__(self, token):
        """Create a simplePBI object to request operations API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token
    def get_operation_state(self, operation_id):
        """Returns the current state of the long running operation
        #### Parameters
        ----
        operation_id: str uuid
            The operation id from the header of a run operation.
        ### Returns
        ----
        Dict:
            A dictionary containing the state of the operation
        """
        headers = {'Content-Type': 'application/json; charset=utf-8', "Authorization": "Bearer {}".format(self.token)}
        res = requests.get("https://api.fabric.microsoft.com/v1/operations/{}".format(operation_id), headers=headers)
        return res.text
    
    def get_operation_result(self, operation_id):
        """Returns the result of the long running operation
        #### Parameters
        ----
        operation_id: str uuid
            The operation id from the header of a run operation.
        ### Returns
        ----
        Dict:
            A dictionary containing the state of the operation
        """
        headers = {'Content-Type': 'application/json; charset=utf-8', "Authorization": "Bearer {}".format(self.token)}
        res = requests.get("https://api.fabric.microsoft.com/v1/operations/{}/result".format(operation_id), headers=headers)
        return res.text
    
class Folders():
    """Simple library to use the Folders api and obtain folder information from it.
    """
    def __init__(self, token):
        """Create a simplePBI object to request operations API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token

    def create_folder(self, workspace_id, folder_name):
        """Creates a folder in the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        folder_name: string
            The folder name.
        ### Returns
        ----
        Response object from requests library. 201 OK
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/folders".format(workspace_id)
            body = {
                "name": folder_name
            }
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.post(url, data=json.dumps(body), headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e) 

    def delete_folder(self, workspace_id, folder_id):
        """Deletes the specified folder in the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        folder_id: str uuid
            The folder id. You can take it from Fabric URL
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/folders/{}".format(workspace_id, folder_id)
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.delete(url, headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def get_folder(self, workspace_id, folder_id):
        """Returns the specified folder in the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        folder_id: str uuid
            The folder id. You can take it from Fabric URL
        ### Returns
        ----
        Dict:
            A dictionary containing a folder in the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/folders/{}".format(workspace_id, folder_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def list_folders(self, workspace_id):
        """Returns a list of folders in the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        ### Returns
        ----
        Dict:
            A dictionary containing all the folders in the workspace.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/folders".format(workspace_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            data = res.json()
            while 'continuationToken' in data and data['continuationToken'] != None:
                url = "https://api.fabric.microsoft.com/v1/workspaces/{}/folders?continuationToken={}".format(workspace_id, data['continuationToken'])
                res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
                res.raise_for_status()
                data.update(res.json())
                data.pop('continuationToken')
            return data
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def update_folder(self, workspace_id, folder_id, displayName):
        """Updates the specified folder in the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        folder_id: str uuid
            The folder id. You can take it from Fabric URL
        displayName: str
            The new display name for the folder.
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/folders/{}".format(workspace_id, folder_id)
            body = {
                "displayName": displayName
            }
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.patch(url, data=json.dumps(body), headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def move_folder(self, workspace_id, folder_id, new_parent_folder_id):
        """Moves the specified folder to a new parent folder in the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        folder_id: str uuid
            The folder id. You can take it from Fabric URL
        new_parent_folder_id: str uuid
            The new parent folder id.
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/workspaces/{}/folders/{}/move".format(workspace_id, folder_id)
            body = {
                "targetFolderId": new_parent_folder_id
            }
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.post(url, data=json.dumps(body), headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

class Connections():
    """Simple library to use the Connections api and obtain connection information from it.
    """
    def __init__(self, token):
        """Create a simplePBI object to request operations API
        Args:
            token: String
                Bearer Token to use the Rest API
        """
        self.token = token

    def create_connection(self, body_connection):
        """Creates a new connection in the specified workspace.
        #### Parameters
        ----
        workspace_id: str uuid
            The workspace id. You can take it from Fabric URL
        body_connection: dict
            The body of the connection to create. Make sure you read https://learn.microsoft.com/en-us/rest/api/fabric/core/connections/create-connection?tabs=HTTP
            Possible types { CreateCloudConnectionRequest, CreateOnPremisesConnectionRequest, CreateVirtualNetworkGatewayConnectionRequest }
        ### Returns
        ----
        Response object from requests library. 201 Created
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/connections"
            body = body_connection
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.post(url, data=json.dumps(body), headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def delete_connection(self, connection_id):
        """Deletes the specified connection.
        #### Parameters
        ----
        connection_id: str uuid
            The connection id. You can take it from Fabric URL
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/connections/{}".format(connection_id)
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.delete(url, headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def get_connection(self, connection_id):
        """Returns the specified connection.
        #### Parameters
        ----
        connection_id: str uuid
            The connection id. You can take it from Fabric URL
        ### Returns
        ----
        Dict:
            A dictionary containing the connection information.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/connections/{}".format(connection_id)
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def list_connections(self):
        """Returns a list of all connections.
        ### Returns
        ----
        List[Dict]:
            A list of dictionaries, each containing information about a connection.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/connections"
            res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
            res.raise_for_status()
            data = res.json()
            while 'continuationToken' in data and data['continuationToken'] != None:
                url = "https://api.fabric.microsoft.com/v1/connections?continuationToken={}".format(data['continuationToken'])
                res = requests.get(url, headers={'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)})
                res.raise_for_status()
                data.update(res.json())
                data.pop('continuationToken')

            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def update_connection(self, connection_id, body_connection):
        """Updates the specified connection.
        #### Parameters
        ----
        connection_id: str uuid
            The connection id. You can take it from Fabric URL
        body_connection: dict
            The body of the connection to update. Make sure you read https://learn.microsoft.com/en-us/rest/api/fabric/core/connections/update-connection?tabs=HTTP
            Possible types { UpdateOnPremisesGatewayConnectionRequest, UpdateOnPremisesGatewayPersonalConnectionRequest, UpdatePersonalCloudConnectionRequest, UpdateShareableCloudConnectionRequest, UpdateVirtualNetworkGatewayConnectionRequest }
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/connections/{}".format(connection_id)
            body = body_connection
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.patch(url, data=json.dumps(body), headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def add_coonection_role_assignment(self, connection_id, principal_id, principal_type, role):
        """Adds a role assignment to the specified connection.
        #### Parameters
        ----
        connection_id: str uuid
            The connection id. You can take it from Fabric URL
        principal_id: str uuid
            The principal id.
        principal_type: string
            The principal type. { "User", "Group", "ServicePrincipal", "ServicePrincipalProfile" }
        role: string
            The role. { "Owner", "User", "UserWithReshare" }
        ### Returns
        ----
        Response object from requests library. 201 Created
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/connections/{}/roleAssignments".format(connection_id)
            body = {
                "principal": {
                    "id": principal_id,
                    "type": principal_type
                },
                "role": role
            }
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.post(url, data=json.dumps(body), headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def list_connection_role_assignments(self, connection_id):
        """Lists all role assignments for the specified connection.
        #### Parameters
        ----
        connection_id: str uuid
            The connection id. You can take it from Fabric URL
        ### Returns
        ----
        List[Dict]:
            A list of dictionaries, each containing information about a role assignment.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/connections/{}/roleAssignments".format(connection_id)
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.get(url, headers=headers)
            res.raise_for_status()
            data = res.json()
            while 'continuationToken' in data and data['continuationToken'] != None:
                url = "https://api.fabric.microsoft.com/v1/connections/{}/roleAssignments?continuationToken={}".format(connection_id, data['continuationToken'])
                res = requests.get(url, headers=headers)
                res.raise_for_status()
                data.update(res.json())
                data.pop('continuationToken')
            return data
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e) 

    def delete_connection_role_assignment(self, connection_id, connection_role_assignment_id):
        """Deletes the specified role assignment from the connection.
        #### Parameters
        ----
        connection_id: str uuid
            The connection id. You can take it from Fabric URL
        connection_role_assignment_id: str uuid
            The role assignment id.
        ### Returns
        ----
        Response object from requests library. 204 No Content
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/connections/{}/roleAssignments/{}".format(connection_id, connection_role_assignment_id)
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.delete(url, headers=headers)
            res.raise_for_status()
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)

    def get_connection_role_assignment(self, connection_id, connection_role_assignment_id):
        """Returns the specified role assignment from the connection.
        #### Parameters
        ----
        connection_id: str uuid
            The connection id. You can take it from Fabric URL
        connection_role_assignment_id: str uuid
            The role assignment id.
        ### Returns
        ----
        Dict:
            A dictionary containing the role assignment information.
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/connections/{}/roleAssignments/{}".format(connection_id, connection_role_assignment_id)
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.get(url, headers=headers)
            res.raise_for_status()
            return res.json()
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e) 

    def update_connection_role_assignment(self, connection_id, connection_role_assignment_id, role):
        """Updates the specified role assignment from the connection.
        #### Parameters
        ----
        connection_id: str uuid
            The connection id. You can take it from Fabric URL
        connection_role_assignment_id: str uuid
            The role assignment id.
        role: string
            The role. { "Owner", "User", "UserWithReshare" }
        ### Returns
        ----
        Response object from requests library. 200 OK
        """
        try:
            url = "https://api.fabric.microsoft.com/v1/connections/{}/roleAssignments/{}".format(connection_id, connection_role_assignment_id)
            body = {
                "role": role
            }
            headers = {'Content-Type': 'application/json', "Authorization": "Bearer {}".format(self.token)}
            res = requests.patch(url, data=json.dumps(body), headers=headers)
            res.raise_for_status()            
            return res
        except requests.exceptions.HTTPError as ex:
            print("HTTP Error: ", ex, "\nText: ", ex.response.text)
        except requests.exceptions.RequestException as e:
            print("Request exception: ", e)